from __future__ import annotations

from jinjarope import textfilters


GIVEN = """def test(sth, averylongvarname, anotherlongvarname, andanother): pass"""
EXPECTED = """\
def test(
    sth,
    averylongvarname,
    anotherlongvarname,
    andanother,
):
    pass
"""


def test_removesuffix():
    assert textfilters.removesuffix("Hello, World!", ", World!") == "Hello"


def test_removeprefix():
    assert textfilters.removeprefix("Hello, World!", "Hello, ") == "World!"


def test_lstrip():
    assert textfilters.lstrip("   Hello, World!  ") == "Hello, World!  "  # noqa: B005


def test_rstrip():
    assert textfilters.rstrip("   Hello, World!  ") == "   Hello, World!"  # noqa: B005


def test_format_code():
    assert textfilters.format_code(GIVEN, line_length=50) == EXPECTED
    assert textfilters.format_code("invalid code!", line_length=50) == "invalid code!"


def test_format_signature():
    def test_function(a, b, c: int = 1, *args, **kwargs):
        pass

    assert (
        textfilters.format_signature(test_function)
        == "(a, b, c: int = 1, *args, **kwargs)"
    )
    assert (
        textfilters.format_signature(test_function, eval_str=False)
        == "(a, b, c: 'int' = 1, *args, **kwargs)"
    )


def test_slugify():
    assert textfilters.slugify("Hello, World!") == "hello__world_"


if __name__ == "__main__":
    code = "def test(sth, fsjkdalfjksdalfjsadk, fjskldjfkdsljf, fsdkjlafjkdsafj): pass"
    text = textfilters.format_code(code, line_length=50)
    print(text)



from __future__ import annotations

from jinjarope import serializefilters


def test_serialize_deserialize():
    text = {"abc": {"def": "ghi"}}
    for fmt in ("yaml", "json", "ini", "toml"):
        assert (
            serializefilters.deserialize(serializefilters.serialize(text, fmt), fmt)
            == text
        )


def test_dig():
    data = {
        "section1": {
            "section2": {
                "section3": "Hello, World!",
            },
        },
    }
    assert (
        serializefilters.dig(data, "section1", "section2", "section3") == "Hello, World!"
    )
    assert serializefilters.dig(data, "section1", "section2", "nonexistent") is None


def test_dig_with_list():
    data = {
        "section1": [
            {"section1": "Wrong one!"},
            {"section2": "Hello, World!"},
        ],
    }
    assert serializefilters.dig(data, "section1", "section2") == "Hello, World!"


def test_dig_with_keep_path():
    data = {
        "section1": {
            "section2": {
                "section3": "Hello, World!",
                "something": "else!",
            },
        },
    }
    assert serializefilters.dig(
        data,
        "section1",
        "section2",
        "section3",
        keep_path=True,
    ) == {
        "section1": {"section2": {"section3": "Hello, World!"}},
    }


def test_dig_with_keep_path_and_list():
    # TODO: this behaviour needs to get implemented correctly.
    data = {
        "section1": [
            {"section1": "Wrong one!"},
            {"section2": "Hello, World!"},
        ],
    }
    assert serializefilters.dig(data, "section1", "section2", keep_path=True) == {
        "section1": {"section2": "Hello, World!"},
    }



from __future__ import annotations

import json
from typing import TYPE_CHECKING

import jinja2
import pytest

from jinjarope import configloaders


if TYPE_CHECKING:
    import pathlib


def test_template_file_loader_json(tmp_path: pathlib.Path):
    d = tmp_path / "example"
    d.mkdir()
    p = d / "template.json"
    p.write_text(json.dumps({"template": "{{ something }}"}))

    loader = configloaders.TemplateFileLoader(p)
    env = jinja2.Environment(loader=loader)
    template = env.get_template("template")
    assert template.render(something="Hello, World!") == "Hello, World!"


def test_template_file_loader_toml(tmp_path: pathlib.Path):
    d = tmp_path / "example"
    d.mkdir()
    p = d / "template.toml"
    p.write_text(r"template = '{{something}}'")

    loader = configloaders.TemplateFileLoader(p)
    env = jinja2.Environment(loader=loader)
    template = env.get_template("template")
    assert template.render(something="Hello, World!") == "Hello, World!"


def test_template_file_loader_not_found(tmp_path: pathlib.Path):
    d = tmp_path / "example"
    d.mkdir()
    p = d / "template.json"
    p.write_text(json.dumps({"template": "{{ something }}"}))

    loader = configloaders.TemplateFileLoader(p)
    env = jinja2.Environment(loader=loader)
    with pytest.raises(jinja2.exceptions.TemplateNotFound):
        env.get_template("nonexistent")


def test_template_file_loader_repr(tmp_path: pathlib.Path):
    d = tmp_path / "example"
    d.mkdir()
    p = d / "template.json"
    p.write_text("{}")

    loader = configloaders.TemplateFileLoader(p)
    assert repr(loader) == f"TemplateFileLoader(path='{p.as_posix()}')"



from __future__ import annotations

import jinjarope


def test_rewrite_loader():
    loader = jinjarope.FileSystemLoader("")
    rewrite_loader = jinjarope.RewriteLoader(loader, lambda p, x: x.replace("C", "XXX"))
    env = jinjarope.Environment()
    env.loader = rewrite_loader
    result = env.render_template("tests/testresources/testfile.jinja")
    assert result.startswith("XXX")



from __future__ import annotations

from collections.abc import Callable, Generator, Iterable, Mapping
import itertools
from typing import Any, TypeVar


T = TypeVar("T")


def pairwise(items: Iterable[T]) -> itertools.pairwise[tuple[T, T]]:
    """Return an iterator of overlapping pairs taken from the input iterator.

    s -> (s0,s1), (s1,s2), (s2, s3), ...

    Arguments:
        items: The items to iter pair-wise
    """
    return itertools.pairwise(items)


def chain(*iterables: Iterable[T]) -> itertools.chain[T]:
    """Chain all given iterators.

    Make an iterator that returns elements from the first iterable until it is
    exhausted, then proceeds to the next iterable, until all of the iterables
    are exhausted. Used for treating consecutive sequences as a single sequence.

    Examples:
        ``` py
        chain('ABC', 'DEF') --> A B C D E F
        ```
    Arguments:
        iterables: The iterables to chain
    """
    return itertools.chain(*iterables)


def product(
    *iterables: Iterable[Any],
    repeat: int = 1,
) -> itertools.product[tuple[Any, ...]]:
    """Cartesian product of input iterables.

    Roughly equivalent to nested for-loops in a generator expression.
    For example, product(A, B) returns the same as ((x,y) for x in A for y in B).

    The nested loops cycle like an odometer with the rightmost element advancing
    on every iteration. This pattern creates a lexicographic ordering so that if
    the input's iterables are sorted, the product tuples are emitted in sorted order.

    To compute the product of an iterable with itself, specify the number of repetitions
    with the optional repeat keyword argument. For example, product(A, repeat=4)
    means the same as product(A, A, A, A).

    Examples:
        ``` py
        product('ABCD', 'xy') --> Ax Ay Bx By Cx Cy Dx Dy
        product(range(2), repeat=3) --> 000 001 010 011 100 101 110 111
        ```

    Arguments:
        iterables: The iterables to create a cartesian product from
        repeat: The amount of repititions
    """
    return itertools.product(*iterables, repeat=repeat)


def repeat(obj: T, times: int | None = None) -> Iterable[T]:
    """Make an iterator that returns object over and over again.

    Runs indefinitely unless the times argument is specified.

    Examples:
        ``` py
        repeat(10, 3) --> 10 10 10
        ```

    Arguments:
        obj: The object to return over and over again
        times: The amount of times to return the object (None means infinite)
    """
    if times:
        return itertools.repeat(obj, times=times)
    return itertools.repeat(obj)


def zip_longest(*iterables: Iterable[Any], fillvalue: Any = None) -> Iterable[Any]:
    """Make an iterator that aggregates elements from each of the iterables.

    If the iterables are of uneven length, missing values are filled-in with fillvalue.
    Iteration continues until the longest iterable is exhausted.

    Examples:
        ``` py
        zip_longest('ABCD', 'xy', fillvalue='-') --> Ax By C- D-
        ```

    Arguments:
        iterables: The iterables to zip
        fillvalue: value to use for filling in case the iterables are of uneven length
    """
    return itertools.zip_longest(*iterables, fillvalue)


def islice(iterable: Iterable[T], *args: int | None) -> itertools.islice[T]:
    """Make an iterator that returns selected elements from the iterable.

    If start is non-zero, then elements from the iterable are skipped until start
    is reached. Afterward, elements are returned consecutively unless step is set
    higher than one which results in items being skipped. If stop is None,
    then iteration continues until the iterator is exhausted, if at all;
    otherwise, it stops at the specified position.

    If start is None, then iteration starts at zero. If step is None,
    then the step defaults to one.

    Unlike regular slicing, islice() does not support negative values
    for start, stop, or step. Can be used to extract related fields from data
    where the internal structure has been flattened (for example, a multi-line report
    may list a name field on every third line).

    Examples:
        ``` py
        islice('ABCDEFG', 2) --> A B
        islice('ABCDEFG', 2, 4) --> C D
        islice('ABCDEFG', 2, None) --> C D E F G
        islice('ABCDEFG', 0, None, 2) --> A C E G
        ```

    Arguments:
        iterable: Iterable to slice
        args: Arguments passed to itertools.islice
    """
    return itertools.islice(iterable, *args)


def do_zip(*items: Iterable[T]) -> zip[tuple[T, ...]]:
    """Zip iterables into a single one.

    Arguments:
        items: The iterables to zip
    """
    return zip(*items)


def reduce_list(items: Iterable[T]) -> list[T]:
    """Reduce duplicate items in a list and preserve order.

    Arguments:
        items: The iterable to recude to a unique-item list
    """
    return list(dict.fromkeys(items))


def flatten_dict(dct: Mapping, sep: str = "/", _parent_key: str = "") -> Mapping:
    """Flatten a nested dictionary to a flat one.

    The individual parts of the "key path" are joined with given separator.

    Arguments:
        dct: The dictionary to flatten
        sep: The separator to use for joining
    """
    items: list[tuple[str, str]] = []
    for k, v in dct.items():
        new_key = _parent_key + sep + k if _parent_key else k
        if isinstance(v, Mapping):
            flattened = flatten_dict(v, _parent_key=new_key, sep=sep)
            items.extend(flattened.items())
        else:
            items.append((new_key, v))
    return dict(items)


def batched(iterable: Iterable[T], n: int) -> Generator[tuple[T, ...], None, None]:
    """Batch data into tuples of length n. The last batch may be shorter.

    Note: this function was added to Py3.12 itertools

    Examples:
        ``` py
        batched('ABCDEFG', 3)  # returns ABC DEF G
        ```

    Arguments:
        iterable: The iterable to yield as batches
        n: The batch size
    """
    if n < 1:
        msg = "n must be at least one"
        raise ValueError(msg)
    it = iter(iterable)
    while batch := tuple(itertools.islice(it, n)):
        yield batch


def natsort(
    val: Iterable[T],
    key: str | Callable[[T], Any] | None = None,
    reverse: bool = False,
    ignore_case: bool = True,
) -> Iterable[T]:
    """Using the natsort package, sort a list naturally.

    i.e. A1, B1, A2, A10 will sort A1, A2, A10, B1.

    Arguments:
        val: the iterable to sort
        key: If str, sort by attribute with given name. If callable, use it as keygetter.
             If None, sort by objects itself
        reverse: Whether to reverse the sort order
        ignore_case: Whether to ignore case for sorting
    """
    from operator import attrgetter

    from natsort import natsorted, ns

    alg = ns.IGNORECASE
    if not ignore_case:
        alg = ns.LOWERCASEFIRST
    key_fn = attrgetter(key) if isinstance(key, str) else key
    return natsorted(val, key=key_fn, reverse=reverse, alg=alg)


def groupby(
    data: Iterable[T],
    key: Callable[[T], Any] | str | None = None,
    *,
    sort_groups: bool = True,
    natural_sort: bool = False,
    reverse: bool = False,
) -> dict[str, list[T]]:
    """Group given iterable using given group function.

    Arguments:
        data: Iterable to group
        key: Sort function or attribute name to use for sorting
        sort_groups: Whether to sort the groups
        natural_sort: Whether to use a natural sort algorithm
        reverse: Whether to reverse the value list
    """
    if key is None:

        def keyfunc(x):
            return x

    elif isinstance(key, str):
        import operator

        keyfunc = operator.attrgetter(key)
    else:
        keyfunc = key
    if sort_groups or natural_sort:
        if natural_sort:
            import natsort

            data = natsort.natsorted(data, key=keyfunc)
        else:
            data = sorted(data, key=keyfunc)
    if reverse:
        data = reversed(list(data))
    return {k: list(g) for k, g in itertools.groupby(data, keyfunc)}


def groupby_first_letter(
    data: Iterable[T],
    keyfunc: Callable[..., Any] | None = None,
) -> dict[str, list[T]]:
    """Group given iterable by first letter.

    Arguments:
        data: Iterable to group
        keyfunc: Optional alternative sort function
    """
    data = sorted(data, key=keyfunc or (lambda x: x))

    def first_letter(x: Any) -> Any:
        return keyfunc(x)[0].upper() if keyfunc else x[0].upper()

    return {k.upper(): list(g) for k, g in itertools.groupby(data, first_letter)}


def do_any(seq: Iterable[Any], attribute: str | None = None) -> bool:
    """Check if at least one of the item in the sequence evaluates to true.

    The `any` builtin as a filter for Jinja templates.

    Arguments:
        seq: An iterable object.
        attribute: The attribute name to use on each object of the iterable.
    """
    if attribute is None:
        return any(seq)
    return any(getattr(i, attribute) for i in seq)



from __future__ import annotations

from jinjarope import iterfilters


def test_reduce_list():
    assert iterfilters.reduce_list([1, 2, 2, 3, 4, 4, 4, 5]) == [1, 2, 3, 4, 5]
    assert iterfilters.reduce_list([]) == []
    assert iterfilters.reduce_list(["a", "b", "b", "c"]) == ["a", "b", "c"]


def test_flatten_dict():
    assert iterfilters.flatten_dict({"a": {"b": {"c": "d"}}}) == {"a/b/c": "d"}
    assert iterfilters.flatten_dict({"a": {"b": "c"}, "d": "e"}) == {"a/b": "c", "d": "e"}
    assert iterfilters.flatten_dict({}) == {}


def test_batched():
    assert list(iterfilters.batched("ABCDEFG", 3)) == [
        ("A", "B", "C"),
        ("D", "E", "F"),
        ("G",),
    ]
    assert list(iterfilters.batched([], 3)) == []
    assert list(iterfilters.batched([1, 2, 3, 4, 5], 2)) == [(1, 2), (3, 4), (5,)]


def test_natsort():
    assert iterfilters.natsort(["A1", "B1", "A2", "A10"], key=lambda x: x) == [
        "A1",
        "A2",
        "A10",
        "B1",
    ]
    assert iterfilters.natsort([], key=lambda x: x) == []
    assert iterfilters.natsort(["B1", "A1"], key=lambda x: x, reverse=True) == [
        "B1",
        "A1",
    ]


def test_do_any():
    assert iterfilters.do_any([True, False, False]) is True
    assert iterfilters.do_any([False, False, False]) is False
    assert iterfilters.do_any([]) is False
    assert iterfilters.do_any([0, 1, 2], attribute="real") is True


def test_groupby_first_letter():
    data = ["apple", "banana", "cherry", "avocado", "carrot", "blueberry"]
    grouped = iterfilters.groupby_first_letter(data)
    assert grouped == {
        "A": ["apple", "avocado"],
        "B": ["banana", "blueberry"],
        "C": ["carrot", "cherry"],
    }


def test_groupby_first_letter_with_empty_data():
    data = []
    grouped = iterfilters.groupby_first_letter(data)
    assert grouped == {}



from __future__ import annotations

from typer.testing import CliRunner

from jinjarope import cli


def test_render_with_cli():
    runner = CliRunner()
    result = runner.invoke(
        cli.cli,
        [
            "-i",
            "file://tests/testresources/testfile.jinja",
            "-j",
            "src/jinjarope/resources/tests.toml",
        ],
        catch_exceptions=False,
    )

    assert result.exit_code == 0
    assert "CONTENT!" in result.output



from __future__ import annotations

import jinja2
import pytest

from jinjarope import jinjaloaderfilesystem


dct = {"home.html": "Home", "about.html": "About", "subfolder/sub.html": "Sub"}


def test_jinja_loader_file_system():
    env = jinja2.Environment(
        loader=jinja2.DictLoader(dct),
    )
    fs = jinjaloaderfilesystem.JinjaLoaderFileSystem(env)

    assert fs.protocol == "jinja"
    assert fs.ls("") == [
        {"name": "subfolder", "type": "directory"},
        {"name": "about.html", "type": "file"},
        {"name": "home.html", "type": "file"},
    ]
    assert fs.ls("", detail=False) == ["subfolder", "about.html", "home.html"]
    assert fs.ls("subfolder/", detail=False) == ["sub.html"]
    assert fs.ls("subfolder/", detail=True) == [{"name": "sub.html", "type": "file"}]
    assert fs.cat("home.html") == b"Home"
    assert fs.cat("about.html") == b"About"
    with pytest.raises(FileNotFoundError):
        fs.cat("nonexistent.html")
    with pytest.raises(FileNotFoundError):
        fs.ls("not-existing-dir")
    fs.env = jinja2.Environment()
    with pytest.raises(FileNotFoundError):
        fs.ls("no-loader-set")
    with pytest.raises(FileNotFoundError):
        fs.open("no-loader-set")
